/*
 
Copyright (C) 2004-2005  Stichting Mathematisch Centrum,
                         Amsterdam, The Netherlands

This file was adapted from an example by Testing Technologies.
 
*/
 
import java.io.IOException;
import java.io.InterruptedIOException;

import java.util.*;

import com.testingtech.ttcn.tci.TciModuleParameterIdImpl;
import com.testingtech.ttcn.tci.TciEncoding;
import com.testingtech.ttcn.tri.TestAdapter;
import com.testingtech.ttcn.tri.TriMessageImpl;
import com.testingtech.ttcn.tri.TriParameterImpl;
import com.testingtech.ttcn.tri.TriStatusImpl;

import org.etsi.ttcn.tci.CharstringValue;
import org.etsi.ttcn.tci.TciCDProvided;
import org.etsi.ttcn.tri.*;


/**
 * TTCN-3 test adapter.
 * All communication between system adapter and the platform adapter
 * is routed through this class. The management part of the adapters is
 * implement in this class. The real work is implemented in separate classes.
 *
 * @author Stefan Blom
 * @version 1.0
 */
public class MyTestAdapter extends TestAdapter implements TriCommunicationSA,
    TriPlatformPA, TciEncoding, TermHandlerComm, Reporter
{

/** Log a message to the TTCN-3 logging interface. */
  public void Log(String message){
	RB.logging.logEvent(System.currentTimeMillis(),this,message);
  }

/** Log a message to the TTCN-3 logging interface and then crash.
 *  TODO: find a more graceful way of dealing with fatal errors.
 */
  public void Fatal(String message){
	RB.logging.logEvent(System.currentTimeMillis(),this,message);
	throw new Error("bailing out.");
  }

/**
 * System under test.
 * Reference to the system adapter for mCRL specification.
 */
  private mCRLadapter SUT = null; // the system under test

/**
 * Port id for outgoing termination messages.
 * Belongs to the termination detection subsystem of simulated time.
 */
  private TriPortId TermPortId;
/**
 * Port id for outgoing termination messages.
 * Belongs to the termination detection subsystem of simulated time.
 */
  private TriComponentId TermComponentId;

/**
 * Create a test adapter.
 */
  public MyTestAdapter() {
    	// invoke the super constructor
	super();
  }

/**
 * Returns a codec.
 */
    public TciCDProvided getCodec(String encodingName)
    {
        if ((encodingName == null) || encodingName.equals(""))
        {
            encodingName = "MyEncoding";
        }

        TciCDProvided codec = super.getCodec(encodingName);

        if (codec != null)
        {
            return codec;
        }

        if (encodingName.equals("MyEncoding"))
        {
            // if with encoding is not set.
            codec = new MyCodec(RB,this);
            codecs.put(encodingName, codec);
        }
        else
        { // the encoding is unknown

            RB.getTciTMProvided().tciError("Unknown decoding " + encodingName);
        }

        // endif encodingName
        return codec;
    }

    /**
     * Execution of the test case will be cancelled by the test manager.
     */
    public void cancel()
    {
	Log("Test execution canceled.");
    }

/**
 * Passing a call from the test system the the SUT.
 */
  public TriStatus triCall(TriComponentId componentId, TriPortId tsiPortId,
      TriAddress sutAddress, TriSignatureId signatureId, TriParameterList parameterList){
            return new TriStatusImpl("triCall: (from: " + componentId +
                ", to: " + tsiPortId + ",SUT address: " + sutAddress +
                ", signature: " + signatureId + ", with paramaters: " +
                parameterList + ") not implemented");
  }

/**
 * Prepare for execution of test case.
 */
    public TriStatus triExecuteTestcase(TriTestCaseId testcase, TriPortIdList tsiList)
    {
	Log("executing "+testcase+" using "+tsiList);
	String LPO = ((CharstringValue) RB.getTciTMProvided().tciGetModulePar(TciModuleParameterIdImpl.valueOf(
          "LPO"))).getString();
	SUT=new mCRLadapter(this,Cte,LPO);
        return new TriStatusImpl();
    }

/**
 * Maps a test component port to a system port.
 */
    public TriStatus triMap(final TriPortId compPortId,
        final TriPortId tsiPortId)
    {
/* Omitting the following call to the superclass will result in crashing the
 * test system!
 */
        TriStatus mapStatus = CsaDef.triMap(compPortId, tsiPortId);
        if (mapStatus.getStatus() != TriStatus.TRI_OK)
        {
            return mapStatus;
        }

/*
 * Below this line the mapping really starts.
 */

	/* The TermOut port is used for sending termination detection messages
	 * from the test adapter to the test engine.
	 */
        if (tsiPortId.getPortName().equals("TermOut")) {
		TermPortId=tsiPortId;
		TermComponentId=compPortId.getComponent();
		return new TriStatusImpl();
	}
	/* The TermIn port is used for sending termination detection messages
	 * from the test engine to the test adapter.
	 */
	if (tsiPortId.getPortName().equals("TermIn")) {
		return new TriStatusImpl();
	}
	/* All other ports are mapped to the mCRL system adapter.
	 */
	return SUT.map(tsiPortId.getPortName(),tsiPortId,compPortId.getComponent());
    }

/**
 * Send a message from the test engine to the test adapter.
 */
    public TriStatus triSend(TriComponentId componentId, TriPortId tsiPortId,
        TriAddress address, TriMessage sendMessage)
    {
	// TermOut is supposed to receive anything.
        if (tsiPortId.getPortName().equals("TermOut")) {
		return new TriStatusImpl("protocol violation");
	}
	// TermIn is routed to the termination detection subsystem.
	if (tsiPortId.getPortName().equals("TermIn")) {
		TermMsgReceived(new TermMsg(sendMessage));
		return new TriStatusImpl();
	}
	// All other messages are routed to the mCRL system adapter.
	return SUT.send(tsiPortId.getPortName(),sendMessage);
    }


/**
 * Unmaps a test component port from a system port.
 */
    public TriStatus triUnmap(TriPortId compPortId, TriPortId tsiPortId)
    {
// Begin of non-omittable code.
        TriStatus mapStatus = CsaDef.triUnmap(compPortId, tsiPortId);
        if (mapStatus.getStatus() != TriStatus.TRI_OK)
        {
            return mapStatus;
        }
// End of non-omittable code.
	// This needs to be implemented for supporting dynamic configurations.
        return mapStatus;
    }



/**
 * Reference to the time manager.
 */
  private TimeManager tm;

/// Forward a termination message from the test adapter to the test engine.
  public void TermMsgSend(TermMsg msg){
	//Log("forwarding termination message from Java to TTCN-3");
	Cte.triEnqueueMsg(TermPortId,null,TermComponentId,msg.toTriMessage());
  }

/// Forward a termination message from the test engineto the test adapter.
  private final void TermMsgReceived(TermMsg msg){
	//Log("forwarding termination message from TTCN-3 to Java");
	NextHandler.TermMsgSend(msg);
  }

/// Next step in the ring along which termination detection messages are sent.
  private TermHandlerComm NextHandler;


/**
 * Reference to the timer unit.
 */
  private TimerUnit tu;

/// Forward timer start command to timer unit.
  public TriStatus triStartTimer(TriTimerId id,TriTimerDuration dur){
	if (tu==null){
		return super.triStartTimer(id,dur);
	} else {
		return tu.triStartTimer(id,dur);
	}
  }

/// Forward timer read command to timer unit.
  public TriStatus triReadTimer(TriTimerId id,TriTimerDuration dur){
	if (tu==null){
		return super.triReadTimer(id,dur);
	} else {
		return tu.triReadTimer(id,dur);
	}
  }

/// Forward timer stop command to timer unit.
  public TriStatus triStopTimer(TriTimerId id){
	if (tu==null){
		return super.triStopTimer(id);
	} else {
		return tu.triStopTimer(id);
	}
  }

/// Forward timer running command to timer unit.
  public TriStatus triTimerRunning(TriTimerId id,TriBoolean b){
	if (tu==null){
		return super.triTimerRunning(id,b);
	} else {
		return tu.triTimerRunning(id,b);
	}
  }

/// Platform adapter reset.
  public TriStatus triPAReset(){
	if (SUT!=null) {
		Log("Kill the mCRL adapter.");
		SUT.halt();
		SUT=null;
	}
	if (tu!=null){
		Log("Kill the timer unit.");
		tu.stop();
		tu=null;
	}
	if (tm!=null){
		Log("Discard the time manager.");
		/* Should deadlocks appear, it might be necessary to make the time
		 * manager active (a thread), which would need killing at this point.
		 */
		tm=null;
	}
	// Do not omit resetting the superclass!
	Log("Continuing platform adapter reset.");
	return super.triPAReset();
  }

/// Route calls to external functions.
  public TriStatus triExternalFunction(TriFunctionId id,TriParameterList pars,TriParameter ret){
	String fun=id.getFunctionName();
	// The function timestart switches from real time to simulated time.
	if (fun=="timestart") {
		tm=new TimeManager(this);
		NextHandler=tm.RegisterHandler(this);
		tu=new TimerUnit(this,tm,Pte);
		SUT.enableSimulatedTime(tm);
		tm.StartProgression();
		return new TriStatusImpl();
	}
	// Not known if anything will break if the superclass is not used.
	return super.triExternalFunction(id,pars,ret);
  }

}



