/*
 
Copyright (C) 2004-2005  Stichting Mathematisch Centrum,
                         Amsterdam, The Netherlands
 
*/

import java.io.IOException;
import java.io.InterruptedIOException;

import java.util.*;

import com.testingtech.ttcn.tci.TciModuleParameterIdImpl;
import com.testingtech.ttcn.tci.TciEncoding;
import com.testingtech.ttcn.tri.TestAdapter;
import com.testingtech.ttcn.tri.TriMessageImpl;
import com.testingtech.ttcn.tri.TriParameterImpl;
import com.testingtech.ttcn.tri.TriStatusImpl;

import org.etsi.ttcn.tci.CharstringValue;
import org.etsi.ttcn.tci.TciCDProvided;
import org.etsi.ttcn.tri.*;


/**
 * Simulated time implementation of TTCN-3 timers.
 *
 * @author Stefan Blom
 * @version 1.0
 */
public class TimerUnit implements Runnable, TimedUnit {

  private TriPlatformTE Pte; // destination for timeouts
  //private TriPortId PortId;
  //private TriComponentId ComponentId;
  private Thread thr; // thread that will generate timeouts
  private Reporter reporter; // for logging messages and errors
  private TermHandler th; // companion termination handler
  private Hashtable BlockedTimers=new Hashtable(); // contains timers with >0 values
  private Hashtable WaitingTimers=new Hashtable(); // contains timers with 0 values
  private Hashtable ExpiredTimers=new Hashtable(); // contains timers with 0 values and timeout sent
  // inactive timers are not stored in any hashtable

  public TimerUnit(
	Reporter reporter,
	TimeManager tm,
	TriPlatformTE Pte
  ){
	this.reporter=reporter;
	this.Pte=Pte;
	th=tm.getTermHandler("Timer Unit");
	tm.add(this);
	thr=new Thread(this);
	thr.start();
  }

  public void stop(){
	thr.interrupt();
	try {
		thr.join();
	} catch (InterruptedException irq){}
  }

  private class IntTimer {
	public int timeleft;
	public int timeset;
	public IntTimer(int t){
		timeset=t;
		timeleft=t;
	}
  }

  public synchronized void tick(){
	/* time is progresssing so we decrease the timeleft of every timer
           and we move the timers whcih have reached 0 to Waiting */
	Iterator i=BlockedTimers.keySet().iterator();
	while(i.hasNext()){
		TriTimerId id=(TriTimerId)i.next();
		IntTimer t=(IntTimer)BlockedTimers.get(id);
		t.timeleft--;
		if (t.timeleft==0){
			reporter.Log("timer "+id+" down to 0");
			i.remove();
			WaitingTimers.put(id,t);
		} else {
			reporter.Log("timer "+id+" has "+t.timeleft+" left");
		}
	}
	/* If there are timers at 0 these need to be expired, which means
           that we must start the next time-slice non-idle. */
	
  }

  public synchronized void next(){
	// Start the new timeslice by signalling our thread to proceed.
      if (!WaitingTimers.isEmpty()) {th.Activate();	notifyAll();}
  }

  public synchronized void run(){
	Iterator i;
	for(;;) {
		/* for all waiting timers, we issue a timeout and we move them
                   to ExpiredTimers */
		i=WaitingTimers.keySet().iterator();
		while(i.hasNext()){
			TriTimerId id=(TriTimerId)i.next();
			IntTimer t=(IntTimer)WaitingTimers.get(id);
			reporter.Log("timer "+id+" expired");
			Pte.triTimeout(id);
			i.remove();
			ExpiredTimers.put(id,t);
		}
		// We removed everything from WaitingTimers, so we know it is empty.
		// If ExpiredTimers is also empty we become Idle.
		// Note that this happens only if we (re)started without waiting timers.
		if (ExpiredTimers.isEmpty()) th.Idle();
		try {
			reporter.Log("waiting to expire timers.");
			// give the others a chance to operate on timers.
			wait();
		} catch (InterruptedException e) {
			reporter.Log("exit due to interrupt.");
			return;
		};
	}
  }

  public synchronized TriStatus triStartTimer(TriTimerId id,TriTimerDuration dur){
	reporter.Log("start timer " + id + " for " + dur.getDuration());
	int val=Math.round((float)dur.getDuration());
	double diff=Math.abs(((double)val)-dur.getDuration());
	if (diff>0.00001) return new TriStatusImpl("timer value not an integer");
	// Start timer overrides all previous settings.
	// So we insert it in the correct table and clear it from other tables.
	if (val==0) {
		BlockedTimers.remove(id);
		WaitingTimers.put(id,new IntTimer(val));
		ExpiredTimers.remove(id);
		// For a timer set to 0 we must ensure that we are in the non-idle
		// state and that the timer will be expired.
		th.Activate(); // Tell TH that we are in the non-idle state
		notifyAll(); // Make sure the timer is expired.
	} else {
		BlockedTimers.put(id,new IntTimer(val));
		WaitingTimers.remove(id);
		ExpiredTimers.remove(id);
		// If we just set the last timer, which was 0 then we now must
		// move to the idle state.
		if (ExpiredTimers.isEmpty()&&ExpiredTimers.isEmpty()) th.Idle();
	}
	return new TriStatusImpl();
  }

  public synchronized TriStatus triReadTimer(TriTimerId id,TriTimerDuration dur){
	// to read a timer we must find it...
	IntTimer tmr=(IntTimer)BlockedTimers.get(id);
	if (tmr==null) tmr=(IntTimer)WaitingTimers.get(id);
	if (tmr==null) tmr=(IntTimer)ExpiredTimers.get(id);
	// ...set the correct duration...
	if (tmr==null) {
		dur.setDuration(0.0);
	} else {
		dur.setDuration(tmr.timeset-tmr.timeleft);
	}
	// ...and return it.
	//reporter.Log("read timer " + id + " gives " + dur.getDuration());
	return new TriStatusImpl();
  }

  public synchronized TriStatus triStopTimer(TriTimerId id){
 	//reporter.Log("stop timer " + id);
	// A stopped timer is in none of the tables.
	BlockedTimers.remove(id);
	WaitingTimers.remove(id);
	ExpiredTimers.remove(id);
	// In case this was the last one at 0 we go idle.
	if (ExpiredTimers.isEmpty()&&ExpiredTimers.isEmpty()) th.Idle();
	return new TriStatusImpl();
  }

  public synchronized TriStatus triTimerRunning(TriTimerId id,TriBoolean b){
	// We assume a timer is running until its timeout receipt has been confirmed!
	boolean running=
		BlockedTimers.containsKey(id)||
		WaitingTimers.containsKey(id)||
		ExpiredTimers.containsKey(id);
	b.setBooleanValue(running);
 	//reporter.Log("timer " + id + " is " + (b.getBooleanValue()?"":"not ") + "running");
	return new TriStatusImpl();
  }

}

